/*
   Copyright The containerd Authors.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package types

import "io"

// CheckpointCreateOptions specifies options for `nerdctl checkpoint create`.
type CheckpointCreateOptions struct {
	Stdout   io.Writer
	GOptions GlobalCommandOptions
	// Leave the container running after checkpointing
	LeaveRunning bool
	// Checkpoint directory
	CheckpointDir string
}

type CheckpointListOptions struct {
	Stdout   io.Writer
	GOptions GlobalCommandOptions
	// Checkpoint directory
	CheckpointDir string
}

// CheckpointRemoveOptions specifies options for `nerdctl checkpoint rm`.
type CheckpointRemoveOptions struct {
	Stdout   io.Writer
	GOptions GlobalCommandOptions
	// Checkpoint directory
	CheckpointDir string
}
type CheckpointSummary struct {
	// Name is the name of the checkpoint.
	Name string
}
